"""
(c) 2022 Jaxcie
This code is licensed under MIT license (see LICENSE for details)
"""


class PyStageLinQError:
    # PyStageLinQ Error codes
    STAGELINQOK = 0

    DISCOVERYTIMEOUT = 100

    INVALIDFRAME = 200
    MAGICFLAGNOTFOUND = 201
    INVALIDLENGTH = 202
    BLOCKLENGTHINCONSISTENT = 203

    INVALIDTOKEN = 300
    INVALIDDISCOVERYDATA = 301
    INVALIDTOKENTYPE = 302
    INVALIDTOKENLENGTH = 303
    INVALIDTOKENVALUE = 304

    CANNOTBINDSOCKET = 400

    SERVICENOTRECOGNIZED = 500
